package coilPred;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.HashSet;

import sequences.AminoAcid;
import sequences.CoilState;
import sequences.CoilType;
import sequences.CoiledCoil;
import sequences.Family;
import sequences.Heptad;
import sequences.Pos;
import sequences.Sequence;
import tables.HeptadPairwiseProbTable;
import tables.HeptadResProbTable;
import tables.PairwiseProbTable;
import tables.ResProbTable;


public class CoilHMM {
	//Code for computing reverse transitions goes here
	public ArrayList<CoilState> HMM_states = new ArrayList<CoilState>();
	ArrayList<CoilState> start_states = new ArrayList<CoilState>();
	ArrayList<CoilState> end_states = new ArrayList<CoilState>();
	HashMap<CoilState,ArrayList<CoilState>> trans = new HashMap<CoilState, ArrayList<CoilState>>();
	HashMap<CoilState,ArrayList<CoilState>> reverse_trans = new HashMap<CoilState, ArrayList<CoilState>>();
	public HashMap<CoilState, HashMap<AminoAcid,Double>> emit_p = new HashMap<CoilState, HashMap<AminoAcid,Double>>();
	public HashMap<Integer, EnumMap<Heptad,EnumMap<AminoAcid,EnumMap<AminoAcid,Double>>>> dimer_corr = new HashMap<Integer, EnumMap<Heptad,EnumMap<AminoAcid,EnumMap<AminoAcid,Double>>>>();
	public HashMap<Integer, EnumMap<Heptad,EnumMap<AminoAcid,EnumMap<AminoAcid,Double>>>> trimer_corr = new HashMap<Integer, EnumMap<Heptad,EnumMap<AminoAcid,EnumMap<AminoAcid,Double>>>>();
	public HashMap<Integer, HashMap<AminoAcid,HashMap<AminoAcid,Double>>> none_corr = new HashMap<Integer, HashMap<AminoAcid,HashMap<AminoAcid,Double>>>();

	//The probability that a given Heptad will output
	//a given AminoAcid:
	public HeptadResProbTable dimerFreqs;
	HeptadResProbTable trimerFreqs;

	//The probability that two specific Heptads of a certain
	//spacing will output two specific AminoAcid residues
	HeptadPairwiseProbTable dimerPairFreqs;
	HeptadPairwiseProbTable trimerPairFreqs;

	//The probability that a non-CoiledCoil region will output
	//a certain AminoAcid residue
	ResProbTable nonCCFreqs;

	//The probability that two residues in a non-CoiledCoil region
	//will output two specific AminoAcids. For example:
	//noCoilPairFreqs.get(3, AminoAcid.A, AminoAcid.C) is the probability
	//that for any two residues in a non-CoiledCoil, the first will be A
	//and the second will be C
	PairwiseProbTable nonCCPairFreqs;	

	//Our parameters...
	double dimerHphobaExp;
	double dimerHphobdExp;
	double trimerHphobaExp;
	double trimerHphobdExp;
	double dimerHphobeExp;
	double dimerHphobgExp;
	double trimerHphobeExp;
	double trimerHphobgExp;

	double dimerNonSingleExp;
	double trimerNonSingleExp;

	double dimerNonGapExp;
	double trimerNonGapExp;

	double[] dimerDP = new double[7];
	double[] trimerDP = new double[7];

	double[] dimerTP = new double[7];
	double[] trimerTP = new double[7];

	double[] dimerD0 = new double[7];
	double[] trimerD0 = new double[7];

	double[] dimerT0 = new double[7];
	double[] trimerT0 = new double[7];

	//double nonGapExp;

	double dimerConst;
	double trimerConst;

	//Constructor - take your 
	public CoilHMM() {
		HashSet<CoilState> dimerStates = new HashSet<CoilState>();
		HashSet<CoilState> trimerStates = new HashSet<CoilState>();
		HashSet<CoilState> endStates = new HashSet<CoilState>();
		//Store all the states in an ArrayList:
		for(Pos p: Pos.values()) {
			//Add the nonCC states:
			{
				//Add the none: position one state
				if(p == Pos.ONE) {
					CoilState s = new CoilState(""+p.toString(),p, CoilType.NONE);
					addState(s);
					endStates.add(s);
				}
			}
			//Add the DIMER states:
			for(Heptad hept: Heptad.values()) {
				CoilState s = new CoilState(""+p.toString()+hept.toString()+"d", hept, p, CoilType.DIMER);
				addState(s);
				dimerStates.add(s);
				if(p.lastState()) {
					endStates.add(s);
				}
			}
			//Add the TRIMER states:
			for(Heptad hept: Heptad.values()) {
				CoilState s = new CoilState(""+p.toString()+hept.toString()+"t", hept, p, CoilType.TRIMER);
				addState(s);
				trimerStates.add(s);
				if(p.lastState()) {
					endStates.add(s);
				}
			}
		}

		for(CoilState s1: HMM_states) {
			for(CoilState s2: HMM_states) {
				if ((s1.getCoilType() == CoilType.NONE) && (s2.getCoilType() == CoilType.NONE)) {
					trans.get(s1).add(s2);
				} else if ((s1.getCoilType() == CoilType.DIMER) && (s2.getCoilType() == CoilType.DIMER) &&
						(s1.position().next() == s2.position()) && (s2.heptad().prev() == s1.heptad())) {
					trans.get(s1).add(s2);
				} else if ((s1.getCoilType() == CoilType.TRIMER) && (s2.getCoilType() == CoilType.TRIMER) &&
						(s1.position().next() == s2.position()) && (s2.heptad().prev() == s1.heptad())) {
					trans.get(s1).add(s2);
				} else if((s1.position().canLeave()) && (s2.position() == Pos.ONE)) {
					trans.get(s1).add(s2);
				} else if((s1.getCoilType() == CoilType.NONE) && (s2.position() == Pos.ONE) && (s2.getCoilType() == CoilType.DIMER)) {
					trans.get(s1).add(s2);
				} else if((s1.getCoilType() == CoilType.NONE) && (s2.position() == Pos.ONE) && (s2.getCoilType() == CoilType.TRIMER)) {
					trans.get(s1).add(s2);
				} else if ((s1.getCoilType() != CoilType.NONE) && (s1.position().canLeave()) && (s2.position() == Pos.ONE) && (s2.getCoilType() == CoilType.NONE)) {
					trans.get(s1).add(s2);
				} else if ((s1.getCoilType() == s2.getCoilType() && s1.position() == Pos.EIGHT && s2.position() == Pos.NINE && s2.heptad().prev() == s1.heptad())) {
					trans.get(s1).add(s2);
				}
			}
		}


		//set the start and end states:
		for(CoilState s1: HMM_states) {
			if((s1.getCoilType() == CoilType.NONE)&& (s1.position() == Pos.ONE)) {
				start_states.add(s1);
			} else if ((s1.position() == Pos.ONE) && (s1.getCoilType() == CoilType.DIMER)) {
				start_states.add(s1);
			} else if ((s1.position() == Pos.ONE) && (s1.getCoilType() == CoilType.TRIMER)) {
				start_states.add(s1);
			}
		}

		for(CoilState s1: HMM_states) {
			if((s1.getCoilType() == CoilType.NONE) && (s1.position() == Pos.ONE)) {
				end_states.add(s1);
			} else if ((s1.position() == Pos.NINE) && (s1.getCoilType() == CoilType.DIMER)) {
				end_states.add(s1);
			} else if ((s1.position() == Pos.NINE) && (s1.getCoilType() == CoilType.TRIMER)) {
				end_states.add(s1);
			}
		}



		setReverseTrans();
	}

	public void addState(CoilState s) {
		HMM_states.add(s);
		trans.put(s, new ArrayList<CoilState>());
		emit_p.put(s, new HashMap<AminoAcid, Double>());		
	}

	public void setProbs(HeptadResProbTable dimerFreqs, HeptadResProbTable trimerFreqs,
			HeptadPairwiseProbTable dimerPairFreqs,	HeptadPairwiseProbTable trimerPairFreqs,
			ResProbTable nonCCFreqs, PairwiseProbTable nonCCPairFreqs) {
		this.dimerFreqs = dimerFreqs;
		this.trimerFreqs = trimerFreqs;
		this.dimerPairFreqs = dimerPairFreqs;
		this.trimerPairFreqs = trimerPairFreqs;
		this.nonCCFreqs = nonCCFreqs;
		this.nonCCPairFreqs = nonCCPairFreqs;
	}

	public void setReverseTrans() {
		for (CoilState s1: trans.keySet()) {
			for (CoilState s2: trans.get(s1)) {
				if (!reverse_trans.keySet().contains(s2)) {
					reverse_trans.put(s2, new ArrayList<CoilState>());
				}
				reverse_trans.get(s2).add(s1);
			}
		}
	}

	public void init(String config_file) throws FileNotFoundException, IOException {
		loadFromFile(new BufferedReader(new FileReader(config_file)));
		set_emit_p();
		set_corr();
	}


	public void load_hmm(String hmm_file) throws FileNotFoundException, IOException {
		loadFromFile(new BufferedReader(new FileReader(hmm_file)));
		System.out.println("Read HMM from file\n");
		set_emit_p();
		set_corr();
	}

	public void write_hmm(String hmm_file) throws IOException {
		System.out.println("About to write\n");
		writeToFile(new BufferedWriter(new FileWriter(hmm_file)));
		System.out.println("Wrote HMM to file\n");
	}

	/*
	public void generate_hmm(String base_dir,String coeffs_file, ArrayList<Family> dimerFams, ArrayList<Family> trimerFams,
			ArrayList<Sequence> negatives) throws FileNotFoundException, InterruptedException, IOException {
		ArrayList<Double> regressionCoeffsList = Regressor.getRegressionCoeffs(this,base_dir,
				coeffs_file, dimerFams, trimerFams, negatives);

		//Generate frequencies:
		this.dimerFreqs = Processor.emissionFrequencies(dimerFams, 1);
		this.trimerFreqs = Processor.emissionFrequencies(trimerFams, 1);
		this.dimerPairFreqs = Processor.allPairwiseResidueFrequencies(dimerFams, 1);
		this.trimerPairFreqs = Processor.allPairwiseResidueFrequencies(trimerFams, 1);
		this.nonCCFreqs = Processor.residueFrequencies(negatives);
		this.nonCCPairFreqs = Processor.allPairwiseResidueFrequencies(negatives);

		//New code for new parameters:
		dimerNonSingleExp = regressionCoeffsList.get(0);
		dimerNonGapExp = regressionCoeffsList.get(1);
		for (int i = 0; i<7; i++) {
			dimerDP[i] = regressionCoeffsList.get(2+i);
		}
		for (int i = 0; i<7; i++) {
			dimerTP[i] = regressionCoeffsList.get(9+i);
		}
		dimerHphobaExp = regressionCoeffsList.get(16);
		dimerHphobdExp = regressionCoeffsList.get(17);
		dimerHphobeExp = regressionCoeffsList.get(18);
		dimerHphobgExp = regressionCoeffsList.get(19);

		for (int i = 0; i<7; i++) {
			dimerD0[i] = regressionCoeffsList.get(20+i);
		}
		for (int i = 0; i<7; i++) {
			dimerT0[i] = regressionCoeffsList.get(27+i);
		}
		dimerConst = regressionCoeffsList.get(34) - 20;

		trimerNonSingleExp = regressionCoeffsList.get(35);
		trimerNonGapExp = regressionCoeffsList.get(36);
		for (int i = 0; i<7; i++) {
			trimerDP[i] = regressionCoeffsList.get(37+i);
		}
		for (int i = 0; i<7; i++) {
			trimerTP[i] = regressionCoeffsList.get(44+i);
		}
		trimerHphobaExp = regressionCoeffsList.get(51);
		trimerHphobdExp = regressionCoeffsList.get(52);
		trimerHphobeExp = regressionCoeffsList.get(53);
		trimerHphobgExp = regressionCoeffsList.get(54);
		for (int i = 0; i<7; i++) {
			trimerD0[i] = regressionCoeffsList.get(55+i);
		}
		for (int i = 0; i<7; i++) {
			trimerT0[i] = regressionCoeffsList.get(62+i);
		}
		trimerConst = regressionCoeffsList.get(69) - 20;
		set_emit_p();
		set_corr();
	}
	*/

	
	public void generate_hmm2(String base_dir,String coeffs_file, ArrayList<Family> dimerFams, ArrayList<Family> trimerFams,
			ArrayList<Sequence> negatives) throws FileNotFoundException, InterruptedException, IOException {
		ArrayList<Double> regressionCoeffsList = Regressor.getRegressionCoeffs2(this,base_dir,
				coeffs_file, dimerFams, trimerFams, negatives);

		//Generate frequencies:
		this.dimerFreqs = Processor.emissionFrequencies(dimerFams, 1);
		this.trimerFreqs = Processor.emissionFrequencies(trimerFams, 1);
		this.dimerPairFreqs = Processor.allPairwiseResidueFrequencies(dimerFams, 1);
		this.trimerPairFreqs = Processor.allPairwiseResidueFrequencies(trimerFams, 1);
		this.nonCCFreqs = Processor.residueFrequencies(negatives);
		this.nonCCPairFreqs = Processor.allPairwiseResidueFrequencies(negatives);

		//New code for new parameters:
		for (int i = 0; i<7; i++) {
			dimerD0[i] = regressionCoeffsList.get(0);
		}
		for (int i = 0; i<7; i++) {
			dimerT0[i] = regressionCoeffsList.get(1);
		}
		dimerNonSingleExp = regressionCoeffsList.get(2);
		for (int i = 0; i<7; i++) {
			dimerDP[i] = regressionCoeffsList.get(3);
		}
		for (int i = 0; i<7; i++) {
			dimerTP[i] = regressionCoeffsList.get(4);
		}
		dimerNonGapExp = regressionCoeffsList.get(5);
		dimerHphobaExp = regressionCoeffsList.get(6);
		dimerHphobdExp = regressionCoeffsList.get(7);

		dimerConst = regressionCoeffsList.get(8) - 20;

		for (int i = 0; i<7; i++) {
			trimerD0[i] = regressionCoeffsList.get(9);
		}
		for (int i = 0; i<7; i++) {
			trimerT0[i] = regressionCoeffsList.get(10);
		}
		trimerNonSingleExp = regressionCoeffsList.get(11);
		for (int i = 0; i<7; i++) {
			trimerDP[i] = regressionCoeffsList.get(12);
		}
		for (int i = 0; i<7; i++) {
			trimerTP[i] = regressionCoeffsList.get(13);
		}
		trimerNonGapExp = regressionCoeffsList.get(14);
		trimerHphobaExp = regressionCoeffsList.get(15);
		trimerHphobdExp = regressionCoeffsList.get(16);

		trimerConst = regressionCoeffsList.get(17) - 20;
		
		set_emit_p();
		set_corr();
	}

	
	public void loadFromFile(BufferedReader b) throws IOException {
		String s;
		s = b.readLine();
		assert s.equals("HMMConfig");
		s = b.readLine();
		assert s.equals("RegressionCoeffs");
		s = b.readLine();
		System.out.println(s);
		dimerNonSingleExp = Double.parseDouble(s);
		dimerNonGapExp = Double.parseDouble(b.readLine());
		for (int i = 0; i<7; i++) {
			dimerDP[i] = Double.parseDouble(b.readLine());
		}
		for (int i = 0; i<7; i++) {
			dimerTP[i] = Double.parseDouble(b.readLine());
		}
		dimerHphobaExp = Double.parseDouble(b.readLine());
		dimerHphobdExp = Double.parseDouble(b.readLine());
		for (int i = 0; i<7; i++) {
			dimerD0[i] = Double.parseDouble(b.readLine());
		}
		for (int i = 0; i<7; i++) {
			dimerT0[i] = Double.parseDouble(b.readLine());
		}
		dimerConst = Double.parseDouble(b.readLine());

		trimerNonSingleExp = Double.parseDouble(b.readLine());
		trimerNonGapExp = Double.parseDouble(b.readLine());
		for (int i = 0; i<7; i++) {
			trimerDP[i] = Double.parseDouble(b.readLine());
		}
		for (int i = 0; i<7; i++) {
			trimerTP[i] = Double.parseDouble(b.readLine());
		}
		trimerHphobaExp = Double.parseDouble(b.readLine());
		trimerHphobdExp = Double.parseDouble(b.readLine());
		for (int i = 0; i<7; i++) {
			trimerD0[i] = Double.parseDouble(b.readLine());
		}
		for (int i = 0; i<7; i++) {
			trimerT0[i] = Double.parseDouble(b.readLine());
		}
		trimerConst = Double.parseDouble(b.readLine());

		dimerFreqs = new HeptadResProbTable(b);
		trimerFreqs = new HeptadResProbTable(b);
		dimerPairFreqs = new HeptadPairwiseProbTable(b);
		trimerPairFreqs = new HeptadPairwiseProbTable(b);
		nonCCFreqs = new ResProbTable(b);
		nonCCPairFreqs = new PairwiseProbTable(b);
	}

	public void writeToFile(BufferedWriter b) throws IOException {
		b.write("HMMConfig\n");
		b.write("RegressionCoeffs\n");
		b.write(String.valueOf(dimerNonSingleExp)+"\n");
		b.write(String.valueOf(dimerNonGapExp)+"\n");
		for (int i = 0; i<7; i++) {
			b.write(String.valueOf(dimerDP[i])+"\n");
		}
		for (int i = 0; i<7; i++) {
			b.write(String.valueOf(dimerTP[i])+"\n");
		}
		b.write(String.valueOf(dimerHphobaExp)+"\n");
		b.write(String.valueOf(dimerHphobdExp)+"\n");
		for (int i = 0; i<7; i++) {
			b.write(String.valueOf(dimerD0[i])+"\n");
		}
		for (int i = 0; i<7; i++) {
			b.write(String.valueOf(dimerT0[i])+"\n");
		}
		b.write(String.valueOf(dimerConst)+"\n");

		b.write(String.valueOf(trimerNonSingleExp)+"\n");
		b.write(String.valueOf(trimerNonGapExp)+"\n");
		for (int i = 0; i<7; i++) {
			b.write(String.valueOf(trimerDP[i])+"\n");
		}
		for (int i = 0; i<7; i++) {
			b.write(String.valueOf(trimerTP[i])+"\n");
		}
		b.write(String.valueOf(trimerHphobaExp)+"\n");
		b.write(String.valueOf(trimerHphobdExp)+"\n");
		for (int i = 0; i<7; i++) {
			b.write(String.valueOf(trimerD0[i])+"\n");
		}
		for (int i = 0; i<7; i++) {
			b.write(String.valueOf(trimerT0[i])+"\n");
		}
		b.write(String.valueOf(trimerConst)+"\n");		

		dimerFreqs.writeObject(b);
		trimerFreqs.writeObject(b);
		dimerPairFreqs.writeObject(b);
		trimerPairFreqs.writeObject(b);
		nonCCFreqs.writeObject(b);
		nonCCPairFreqs.writeObject(b);		
	}

	/**
	 * Initialize the emit_p table, which includes all effects except for the
	 * dimer/trimer specific pairwise ones.
	 * @return
	 */
	public void set_emit_p() {
		HashMap<CoilState, HashMap<AminoAcid,Double>> T = new HashMap<CoilState, HashMap<AminoAcid,Double>>();
		for (CoilState state: HMM_states) {
			T.put(state, new HashMap<AminoAcid,Double>());
			Heptad h = state.hept;
			CoilType type = state.type;

			for (AminoAcid aa: AminoAcid.values()) {
				double p = 1;
				if (type == CoilType.DIMER) {
					switch (h) {
					case a: {
						p *= Math.exp(aa.hydrophobicity() * dimerHphobaExp);
						p *= Math.pow(dimerFreqs.get(h, aa),dimerD0[0]);
						p *= Math.pow(dimerFreqs.get(h, aa),dimerT0[0]);
						break;
					}
					case b: {
						p *= Math.pow(dimerFreqs.get(h, aa),dimerD0[1]);
						p *= Math.pow(dimerFreqs.get(h, aa),dimerT0[1]);
						break;
					}
					case c: {
						p *= Math.pow(dimerFreqs.get(h, aa),dimerD0[2]);
						p *= Math.pow(dimerFreqs.get(h, aa),dimerT0[2]);
						break;
					}
					case d: {
						p *= Math.exp(aa.hydrophobicity() * dimerHphobdExp);
						p *= Math.pow(dimerFreqs.get(h, aa),dimerD0[3]);
						p *= Math.pow(dimerFreqs.get(h, aa),dimerT0[3]);
						break;
					}
					case e: {
						p *= Math.exp(aa.hydrophobicity() * dimerHphobeExp);
						p *= Math.pow(dimerFreqs.get(h, aa),dimerD0[4]);
						p *= Math.pow(dimerFreqs.get(h, aa),dimerT0[4]);
						break;
					}
					case f: {
						p *= Math.pow(dimerFreqs.get(h, aa),dimerD0[5]);
						p *= Math.pow(dimerFreqs.get(h, aa),dimerT0[5]);
						break;
					}
					case g: {
						p *= Math.exp(aa.hydrophobicity() * dimerHphobgExp);
						p *= Math.pow(dimerFreqs.get(h, aa),dimerD0[6]);
						p *= Math.pow(dimerFreqs.get(h, aa),dimerT0[6]);
						break;
					}
					}
					//We include the constant from the logistic regression
					//here
					if (state.pos == Pos.ONE) {
						p *= Math.exp(dimerConst);
					}
					p *= Math.pow(nonCCFreqs.get(aa),dimerNonSingleExp);
				} else if (type == CoilType.TRIMER) {
					switch (h) {
					case a: {
						p *= Math.exp(aa.hydrophobicity() * trimerHphobaExp);
						p *= Math.pow(dimerFreqs.get(h, aa),trimerD0[0]);
						p *= Math.pow(dimerFreqs.get(h, aa),trimerT0[0]);						
						break;
					}
					case b: {
						p *= Math.pow(dimerFreqs.get(h, aa),trimerD0[1]);
						p *= Math.pow(dimerFreqs.get(h, aa),trimerT0[1]);
						break;
					}
					case c: {
						p *= Math.pow(dimerFreqs.get(h, aa),trimerD0[2]);
						p *= Math.pow(dimerFreqs.get(h, aa),trimerT0[2]);
						break;
					}
					case d: {
						p *= Math.exp(aa.hydrophobicity() * trimerHphobdExp);
						p *= Math.pow(dimerFreqs.get(h, aa),trimerD0[3]);
						p *= Math.pow(dimerFreqs.get(h, aa),trimerT0[3]);
						break;
					}
					case e: {
						p *= Math.exp(aa.hydrophobicity() * trimerHphobeExp);
						p *= Math.pow(dimerFreqs.get(h, aa),trimerD0[4]);
						p *= Math.pow(dimerFreqs.get(h, aa),trimerT0[4]);
						break;
					}
					case f: {
						p *= Math.pow(dimerFreqs.get(h, aa),trimerD0[5]);
						p *= Math.pow(dimerFreqs.get(h, aa),trimerT0[5]);
						break;
					}
					case g: {
						p *= Math.exp(aa.hydrophobicity() * trimerHphobgExp);
						p *= Math.pow(dimerFreqs.get(h, aa),trimerD0[6]);
						p *= Math.pow(dimerFreqs.get(h, aa),trimerT0[6]);
						break;
					}
					}
					if (state.pos == Pos.ONE) {
						p *= Math.exp(trimerConst);
					}
					p *= Math.pow(nonCCFreqs.get(aa),trimerNonSingleExp);
				} else { //CoilType == NONE  don't do anything
				}
				T.get(state).put(aa, p);
			}
		}
		emit_p = T;
	}

	public void set_corr() {
		HashMap<Integer, EnumMap<Heptad,EnumMap<AminoAcid,EnumMap<AminoAcid,Double>>>> T  = new HashMap<Integer, EnumMap<Heptad,EnumMap<AminoAcid,EnumMap<AminoAcid,Double>>>>();
		for (int dist=1; dist<=7; dist++) {
			T.put(dist, new EnumMap<Heptad, EnumMap<AminoAcid, EnumMap<AminoAcid,Double>>>(Heptad.class));
			for (Heptad h: Heptad.values()) {
				T.get(dist).put(h, new EnumMap<AminoAcid, EnumMap<AminoAcid,Double>>(AminoAcid.class));
				for (AminoAcid aaPrime: AminoAcid.values()) {
					T.get(dist).get(h).put(aaPrime, new EnumMap<AminoAcid,Double>(AminoAcid.class));
					for (AminoAcid aa: AminoAcid.values()) {
						double val = 1.0;
						Heptad hPrime = h.prev(dist);
						double distDProb = dimerPairFreqs.get(hPrime, h, dist, aaPrime, aa) / (dimerFreqs.get(hPrime, aaPrime) * dimerFreqs.get(h, aa));
						val *= Math.pow(distDProb,dimerDP[dist-1]);
						double distTProb = trimerPairFreqs.get(hPrime, h, dist, aaPrime, aa) / (trimerFreqs.get(hPrime, aaPrime) * trimerFreqs.get(h, aa));
						val *= Math.pow(distTProb,dimerTP[dist-1]);
						//Add the noncoil correlation regardless:
						double distNProb = nonCCPairFreqs.get(dist, aaPrime, aa) / (nonCCFreqs.get(aaPrime) * nonCCFreqs.get(aa));
						val *= Math.pow(distNProb, dimerNonGapExp);
						T.get(dist).get(h).get(aaPrime).put(aa,val);
					}
				}
			}
		}
		dimer_corr = T;

		//Trimer_corr
		T  = new HashMap<Integer, EnumMap<Heptad,EnumMap<AminoAcid,EnumMap<AminoAcid,Double>>>>();
		for (int dist=1; dist<=7; dist++) {
			T.put(dist, new EnumMap<Heptad, EnumMap<AminoAcid, EnumMap<AminoAcid,Double>>>(Heptad.class));
			for (Heptad h: Heptad.values()) {
				T.get(dist).put(h, new EnumMap<AminoAcid, EnumMap<AminoAcid,Double>>(AminoAcid.class));
				for (AminoAcid aaPrime: AminoAcid.values()) {
					T.get(dist).get(h).put(aaPrime, new EnumMap<AminoAcid,Double>(AminoAcid.class));
					for (AminoAcid aa: AminoAcid.values()) {
						double val = 1.0;
						Heptad hPrime = h.prev(dist);
						double distTProb = trimerPairFreqs.get(hPrime, h, dist, aaPrime, aa) / (trimerFreqs.get(hPrime, aaPrime) * trimerFreqs.get(h, aa));
						val *= Math.pow(distTProb,trimerTP[dist-1]);
						double distDProb = dimerPairFreqs.get(hPrime, h, dist, aaPrime, aa) / (dimerFreqs.get(hPrime, aaPrime) * dimerFreqs.get(h, aa));
						val *= Math.pow(distDProb,trimerDP[dist-1]);
						double distNProb = nonCCPairFreqs.get(dist, aaPrime, aa) / (nonCCFreqs.get(aaPrime) * nonCCFreqs.get(aa));
						val *= Math.pow(distNProb, trimerNonGapExp);
						T.get(dist).get(h).get(aaPrime).put(aa,val);
					}
				}
			}
		}
		trimer_corr = T;
	}
	
	public String scoreCoilEnergy(CoiledCoil cc) {

		HashMap<Integer, Double> nonCCGapEnergies = new HashMap<Integer, Double>();

		HashMap<Integer, Double> dimerGapEnergies = new HashMap<Integer, Double>();
		HashMap<Integer, Double> trimerGapEnergies = new HashMap<Integer, Double>();

		double nonCCSingleEnergy = 0.0;

		HashMap<Heptad, Double> dimerSingleEnergies = new HashMap<Heptad, Double>();
		HashMap<Heptad, Double> trimerSingleEnergies = new HashMap<Heptad, Double>();

		HashMap<Heptad, Double> hydrophobicities = new HashMap<Heptad, Double>();
		HashMap<Heptad, Double> charges = new HashMap<Heptad, Double>();
		HashMap<Heptad, Double> smalls = new HashMap<Heptad, Double>();
		HashMap<Heptad, Double> tinies = new HashMap<Heptad, Double>();
		HashMap<Heptad, Double> branched = new HashMap<Heptad, Double>();

		//Testing: help with convexity?
		double nonCCSingleEnergy2 = 0.0;
		double nonCCSingleEnergy5 = 0.0;

		int[] gapsUsed = {1,2,3,4,5,6,7};

		//Initialize the HashMaps
		for (int i: gapsUsed) {
			nonCCGapEnergies.put(i, 0.0);
			dimerGapEnergies.put(i, 0.0);
			trimerGapEnergies.put(i, 0.0);
			for (Heptad hept: Heptad.values()) {
				dimerSingleEnergies.put(hept, 0.0);

				trimerSingleEnergies.put(hept, 0.0);
				hydrophobicities.put(hept, 0.0);
				charges.put(hept, 0.0);
				smalls.put(hept, 0.0);
				tinies.put(hept, 0.0);
				branched.put(hept, 0.0);
			}
		}


		//Compute the hydrophobicities scores:
		for(int i=0; i<cc.length(); i++) {
			hydrophobicities.put(cc.heptad(i), hydrophobicities.get(cc.heptad(i))+cc.residue(i).hydrophobicity());
		}


		//Compute the charge scores:
		for(int i=0; i<cc.length(); i++) {
			charges.put(cc.heptad(i), charges.get(cc.heptad(i))+cc.residue(i).charge());
		}

		//Compute the small
		for(int i=0; i<cc.length(); i++) {
			smalls.put(cc.heptad(i), smalls.get(cc.heptad(i))+cc.residue(i).small());
		}

		//Compute the tinies
		for(int i=0; i<cc.length(); i++) {
			tinies.put(cc.heptad(i), tinies.get(cc.heptad(i))+cc.residue(i).tiny());
		}

		//Compute the branched
		for(int i=0; i<cc.length(); i++) {
			branched.put(cc.heptad(i), branched.get(cc.heptad(i))+cc.residue(i).branchedSideChain());
		}


		//Compute the singleEnergies scores:
		for(int i=0; i<cc.length(); i++) {
			nonCCSingleEnergy += Math.log(nonCCFreqs.get(cc.residue(i)));
			nonCCSingleEnergy2 += Math.pow(Math.log(nonCCFreqs.get(cc.residue(i))), 2.0);
			nonCCSingleEnergy5 += Math.pow(Math.log(nonCCFreqs.get(cc.residue(i))), .5);
			dimerSingleEnergies.put(cc.heptad(i), dimerSingleEnergies.get(cc.heptad(i)) + 
					Math.log(dimerFreqs.get(cc.heptad(i), cc.residue(i))));
			//			shortDimerSingleEnergies.put(cc.heptad(i), shortDimerSingleEnergies.get(cc.heptad(i)) + 
			//					Math.log(shortDimerFreqs.get(cc.heptad(i), cc.residue(i))));	
			trimerSingleEnergies.put(cc.heptad(i), trimerSingleEnergies.get(cc.heptad(i)) + 
					Math.log(trimerFreqs.get(cc.heptad(i), cc.residue(i))));
		}
		//Compute the gapEnergies scores:
		for(int gap: gapsUsed) {
			for(int i=gap; i<cc.length(); i++) {
				double dimerprob1 = dimerPairFreqs.get(cc.heptad(i-gap), cc.heptad(i), gap, cc.residue(i-gap), cc.residue(i));
				double dimerprob2 = dimerFreqs.get(cc.heptad(i), cc.residue(i)) * dimerFreqs.get(cc.heptad(i-gap), cc.residue(i-gap));
				double trimerprob1 = trimerPairFreqs.get(cc.heptad(i-gap), cc.heptad(i), gap, cc.residue(i-gap), cc.residue(i));
				double trimerprob2 = trimerFreqs.get(cc.heptad(i), cc.residue(i)) * trimerFreqs.get(cc.heptad(i-gap), cc.residue(i-gap));
				double nonprob1 = nonCCPairFreqs.get(gap, cc.residue(i-gap), cc.residue(i));
				double nonprob2 = nonCCFreqs.get(cc.residue(i)) * nonCCFreqs.get(cc.residue(i-gap));
				nonCCGapEnergies.put(gap, nonCCGapEnergies.get(gap) + Math.log(nonprob1/nonprob2));
				dimerGapEnergies.put(gap,
						dimerGapEnergies.get(gap) + Math.log(dimerprob1/dimerprob2));
				trimerGapEnergies.put(gap,
						trimerGapEnergies.get(gap) + Math.log(trimerprob1/trimerprob2));
			}
		}

		//Compute so useful values:
		double dimerSingleEnergy = 0.0;
		double trimerSingleEnergy = 0.0;
		double nonCCtotalGapEnergy = 0.0;
		double dimerTotalGapEnergy = 0.0;
		double trimerTotalGapEnergy = 0.0;


		for(int i: gapsUsed) {
			dimerTotalGapEnergy += dimerGapEnergies.get(i);
			trimerTotalGapEnergy += trimerGapEnergies.get(i);
			nonCCtotalGapEnergy += nonCCGapEnergies.get(i);
		}
		for(Heptad h: Heptad.values()) {
			dimerSingleEnergy += dimerSingleEnergies.get(h);
			trimerSingleEnergy += trimerSingleEnergies.get(h);
		}


		//Compute the output string:
		String result = "" + nonCCSingleEnergy +"\t" + dimerSingleEnergy +"\t" + trimerSingleEnergy + "\t" +
		nonCCtotalGapEnergy +"\t" + dimerTotalGapEnergy +"\t" +trimerTotalGapEnergy;
		for (Heptad hept: Heptad.values()) {
			result += "\t" + dimerSingleEnergies.get(hept);
			result += "\t" + trimerSingleEnergies.get(hept);
		}
		for (int i: gapsUsed) {
			result += "\t" + nonCCGapEnergies.get(i);
			result += "\t" + dimerGapEnergies.get(i);
			result += "\t" + trimerGapEnergies.get(i);
		}
		for (Heptad h: Heptad.values()) {
			result += "\t" + hydrophobicities.get(h);
		}
		for (Heptad h: Heptad.values()) {
			result += "\t" + charges.get(h);
		}
		for (Heptad h: Heptad.values()) {
			result += "\t" + smalls.get(h);
		}
		for (Heptad h: Heptad.values()) {
			result += "\t" + tinies.get(h);
		}
		for (Heptad h: Heptad.values()) {
			result += "\t" + branched.get(h);
		}
		result += "\t" + nonCCSingleEnergy2;
		result += "\t" + nonCCSingleEnergy5;
		return result;
	}

	
	public String scoreCoilEnergy2(CoiledCoil cc) {

		HashMap<Integer, Double> nonCCGapEnergies = new HashMap<Integer, Double>();
		HashMap<Integer, Double> dimerGapEnergies = new HashMap<Integer, Double>();
		HashMap<Integer, Double> trimerGapEnergies = new HashMap<Integer, Double>();

		double nonCCSingleEnergy = 0.0;
		HashMap<Heptad, Double> dimerSingleEnergies = new HashMap<Heptad, Double>();
		HashMap<Heptad, Double> trimerSingleEnergies = new HashMap<Heptad, Double>();

		HashMap<Heptad, Double> hydrophobicities = new HashMap<Heptad, Double>();

		int[] gapsUsed = {1,2,3,4,5,6,7};

		//Initialize the HashMaps
		for (int i: gapsUsed) {
			nonCCGapEnergies.put(i, 0.0);
			dimerGapEnergies.put(i, 0.0);
			trimerGapEnergies.put(i, 0.0);
			for (Heptad hept: Heptad.values()) {
				dimerSingleEnergies.put(hept, 0.0);
				trimerSingleEnergies.put(hept, 0.0);
				hydrophobicities.put(hept, 0.0);
			}
		}

		//Compute the hydrophobicities scores:
		for(int i=0; i<cc.length(); i++) {
			hydrophobicities.put(cc.heptad(i), hydrophobicities.get(cc.heptad(i))+cc.residue(i).hydrophobicity());
		}

		//Compute the singleEnergies scores:
		for(int i=0; i<cc.length(); i++) {
			nonCCSingleEnergy += Math.log(nonCCFreqs.get(cc.residue(i)));
			dimerSingleEnergies.put(cc.heptad(i), dimerSingleEnergies.get(cc.heptad(i)) + 
					Math.log(dimerFreqs.get(cc.heptad(i), cc.residue(i))));
			trimerSingleEnergies.put(cc.heptad(i), trimerSingleEnergies.get(cc.heptad(i)) + 
					Math.log(trimerFreqs.get(cc.heptad(i), cc.residue(i))));
		}

		//Compute the gapEnergies scores:
		for(int gap: gapsUsed) {
			for(int i=gap; i<cc.length(); i++) {
				double dimerprob1 = dimerPairFreqs.get(cc.heptad(i-gap), cc.heptad(i), gap, cc.residue(i-gap), cc.residue(i));
				double dimerprob2 = dimerFreqs.get(cc.heptad(i), cc.residue(i)) * dimerFreqs.get(cc.heptad(i-gap), cc.residue(i-gap));
				double trimerprob1 = trimerPairFreqs.get(cc.heptad(i-gap), cc.heptad(i), gap, cc.residue(i-gap), cc.residue(i));
				double trimerprob2 = trimerFreqs.get(cc.heptad(i), cc.residue(i)) * trimerFreqs.get(cc.heptad(i-gap), cc.residue(i-gap));
				double nonprob1 = nonCCPairFreqs.get(gap, cc.residue(i-gap), cc.residue(i));
				double nonprob2 = nonCCFreqs.get(cc.residue(i)) * nonCCFreqs.get(cc.residue(i-gap));
				nonCCGapEnergies.put(gap, nonCCGapEnergies.get(gap) + Math.log(nonprob1/nonprob2));
				dimerGapEnergies.put(gap,
						dimerGapEnergies.get(gap) + Math.log(dimerprob1/dimerprob2));
				trimerGapEnergies.put(gap,
						trimerGapEnergies.get(gap) + Math.log(trimerprob1/trimerprob2));
			}
		}

		//Compute some useful values:
		double dimerSingleEnergy = 0.0;
		double trimerSingleEnergy = 0.0;
		double nonCCtotalGapEnergy = 0.0;
		double dimerTotalGapEnergy = 0.0;
		double trimerTotalGapEnergy = 0.0;


		for(int i: gapsUsed) {
			dimerTotalGapEnergy += dimerGapEnergies.get(i);
			trimerTotalGapEnergy += trimerGapEnergies.get(i);
			nonCCtotalGapEnergy += nonCCGapEnergies.get(i);
		}
		for(Heptad h: Heptad.values()) {
			dimerSingleEnergy += dimerSingleEnergies.get(h);
			trimerSingleEnergy += trimerSingleEnergies.get(h);
		}


		//Compute the output string:
		String result = "" + nonCCSingleEnergy +"\t" + dimerSingleEnergy +"\t" + trimerSingleEnergy + "\t" +
		nonCCtotalGapEnergy +"\t" + dimerTotalGapEnergy +"\t" +trimerTotalGapEnergy;
		result += "\t" + hydrophobicities.get(Heptad.a);
		result += "\t" + hydrophobicities.get(Heptad.d);
		return result;
	}
	
}