package coilPred;


import java.util.ArrayList;
import java.util.HashMap;

import sequences.AminoAcid;
import sequences.CoilResidue;
import sequences.CoilState;
import sequences.CoilType;
import sequences.CoiledCoil;
import sequences.Family;
import sequences.Heptad;
import sequences.Residue;
import sequences.Sequence;
import tables.HMMPosProbTable;
import tables.HeptadPairwiseProbTable;
import tables.HeptadResProbTable;
import tables.PairwiseProbTable;
import tables.ResProbTable;

public class Processor {
	/**
	 * Returns the total residue frequencies in a list of coiled coils
	 * @param coilList
	 * @return
	 */
	public static ResProbTable residueFrequencies(ArrayList<Sequence> proteinList) {
		ResProbTable table = new ResProbTable();
		//Initialize with a Bernoulli conjugate prior. We assume we have measured each aa once:
		for(AminoAcid aa: AminoAcid.standardValues()) {
			table.add(aa, 1.0);
		}

		for(Sequence s: proteinList) {
			for(Residue residue: s.getResidueList()) {
				if(residue.getAA().nonStandard()) {continue;}
				table.add(residue.getAA(), 1.0);
			}
		}

		return table;
	}


	/**
	 * Returns the Emission Frequencies in a list of coiled coils
	 * @param familyList
	 * @return Table with heptad, aminoacid frequency probabilities
	 */
	public static HeptadResProbTable emissionFrequencies(ArrayList<Family> familyList, double priorWeights) {
		HeptadResProbTable table = new HeptadResProbTable();
		//initialize with conjugate prior - assume priorWeights observation of each:
		for(Heptad h: Heptad.values()) {
			for(AminoAcid aa: AminoAcid.standardValues()) {
				table.add(h, aa, priorWeights);
			}
		}

		//Compute:
		for(Family f: familyList) {
			for(CoiledCoil cc: f.famToCoils()) {
				for(CoilResidue residue: cc.residueList()) {
					if(residue.getAA().nonStandard()) {continue;}
					table.add(residue.getHept(), residue.getAA(), 1.0);
				}
			}
		}
		return table;
	}

	public static HeptadResProbTable emissionFrequencies(HMMPosProbTable<CoilState> HMMstate, ArrayList<AminoAcid> observedList, CoilType ccType) {
		HeptadResProbTable result = new HeptadResProbTable();
		//Go through all the times and states of the HMMstate to fill up the result table:
		for(int i=0; i<observedList.size(); i++) {
			for(CoilState state: HMMstate.get(i).keySet()) {
				if(state.getCoilType() == ccType) {
					result.add(state.heptad(), observedList.get(i), HMMstate.get(i).get(state));
				}
			}
		}
		return result;
	}



	private static HashMap<Family, Double> getFamilyWeights(ArrayList<Family> familyList) {
		HashMap<Family, Double> result = new HashMap<Family, Double>();
		for (Family f: familyList) {
			//			result.put(f, Math.pow(f.famToCoils().size(), -1));
			result.put(f, 1.0);
		}
		return result;
	}



	public static HeptadPairwiseProbTable allPairwiseResidueFrequencies(ArrayList<Family> famList, double priorWeights) {
		//Add all the needed values one at a time - at the moment, only use the
		//distances 1, 2, 4, so:
		int[] array = {1,2,3,4,5,6,7};
		HeptadPairwiseProbTable table = new HeptadPairwiseProbTable(array);
		for (int i: array) {
			for(Heptad h1: Heptad.values()) {
				for(Heptad h2: Heptad.values()) {
					table.add(h1, h2, pairwiseResidueFrequencies(famList, h1, h2, i, priorWeights));
				}
			}
		}
		return table;
	}

	/**
	 * Return the pairwise frequencies for which amino acids will correspond to two heptads of a certain spacing
	 * @param coilList
	 * @param aa1
	 * @param aa2
	 * @param spacing
	 * @return the pairwise frequency of a pair of heptads with the given spacing. For example, inputting
	 * pairwiseResidueFrequencies(coilList, a, d, 3) returns the pairwise frequencies of amino acids
	 * that are in the 'a' and 'd' positions in the heptad, with the one in the 'a' position 3 positions
	 * before the one in the 'd' position (i.e. they are in the same heptad and there are no skips between them)
	 * 
	 * Format of returned value: HashMap<residue in first position, HashMap<residue in second position, probability>>
	 */
	public static 
	//HashMap<AminoAcid, HashMap<AminoAcid, Double>> 
	PairwiseProbTable
	pairwiseResidueFrequencies(ArrayList<CoiledCoil> coilList, Heptad h1, Heptad h2, int spacing) {
		PairwiseProbTable table = new PairwiseProbTable(spacing);
		//initialize the occurrenceTable with conjugate prior: we've seen each pair of AAs once for each h1,h2,spacing
		for(AminoAcid aa1: AminoAcid.standardValues()) {
			for (AminoAcid aa2: AminoAcid.standardValues()) {
				table.add(spacing, aa1, aa2, 1.0);
			}
		}
		for(CoiledCoil coil: coilList) {
			for(int i = 0; i<coil.residueList().size(); i++) {
				if(coil.residueList().size() <= (i + spacing)) {continue;}
				if(coil.residue(i).nonStandard()) {continue;}
				if(coil.residue(i+spacing).nonStandard()) {continue;}
				//First check to see if the (i, i+spacing) pair in the coiled coil has a suitable pair of heptad positions
				if ((coil.heptad(i).equals(h1)) && 
						(coil.heptad(i+spacing).equals(h2))) {
					//if so, update the occurrenceTable:
					table.add(spacing,coil.residue(i),coil.residue(i+spacing),1.0);
				}
			}
		}
		return table;
	}

	public static 
	//HashMap<AminoAcid, HashMap<AminoAcid, Double>> 
	PairwiseProbTable
	pairwiseResidueFrequencies(ArrayList<Family> famList, Heptad h1, Heptad h2, int spacing, double priorWeights) {
		PairwiseProbTable table = new PairwiseProbTable(spacing);
		HashMap<Family, Double> familyWeights = null;

		//initialize the occurrenceTable with conjugate prior: we've seen each pair of AAs once for each h1,h2,spacing
		for(AminoAcid aa1: AminoAcid.standardValues()) {
			for (AminoAcid aa2: AminoAcid.standardValues()) {
				table.add(spacing, aa1, aa2, priorWeights);
			}
		}

		for(Family fam: famList) {
			for(CoiledCoil coil: fam.famToCoils()) {
				for(int i = 0; i<coil.residueList().size(); i++) {
					if(coil.residueList().size() <= (i + spacing)) {continue;}
					if(coil.residue(i).nonStandard()) {continue;}
					if(coil.residue(i+spacing).nonStandard()) {continue;}
					//First check to see if the (i, i+spacing) pair in the coiled coil has a suitable pair of heptad positions
					if ((coil.heptad(i).equals(h1)) && 
							(coil.heptad(i+spacing).equals(h2))) {
						//if so, update the occurrenceTable:
						table.add(spacing,coil.residue(i),coil.residue(i+spacing),1.0);
					}
				}
			}
		}
		return table;
	}


	public static PairwiseProbTable allPairwiseResidueFrequencies(ArrayList<Sequence> proteinList) {
		int[] array = {1,2,3,4,5,6,7};
		PairwiseProbTable result = new PairwiseProbTable(array);
		for(int spacing: array) {
			result.put(pairwiseResidueFrequencies(proteinList, spacing));
		}
		return result;
	}

	public static PairwiseProbTable pairwiseResidueFrequencies(ArrayList<Sequence> proteinList, int spacing) {
		PairwiseProbTable table = new PairwiseProbTable(spacing);
		//initialize the occurrenceTable with a conjugate prior: we've seen each pair of AAs once
		for(AminoAcid aa1: AminoAcid.standardValues()) {
			for (AminoAcid aa2: AminoAcid.standardValues()) {
				table.add(spacing, aa1, aa2, 1.0);
			}
		}		

		for(Sequence s: proteinList) {
			for(int i = 0; i<s.length(); i++) {
				if (s.length() > (i + spacing)) {
					//Update the occurrenceTable:
					table.add(spacing,s.getAA(i),s.getAA(i+spacing),1.0);
				}
			}
		}
		return table;
	}



	public static ResProbTable outOfCoilProbs(ArrayList<Family> coilList) {
		ResProbTable table = new ResProbTable();
		//Initialize with a Bernoulli conjugate prior. We assume we have measured each aa once:
		for(AminoAcid aa: AminoAcid.standardValues()) {
			table.add(aa, 1.0);
		}
		for(Family fam: coilList) {
			for(Sequence seq: fam.getSeqs()) {
				for(Residue r: seq.toNonCoilResidues().getResidueList()) {
					if(r.getAA().nonStandard()) {continue;}
					table.add(r.getAA(), 1.0);
				}
			}
		}
		return table;
	}

	/**
	 * Return a map from lengths of coils to the number of coils with that length
	 * @param coilList
	 * @return
	 */
	public static HashMap<Integer, Integer> lengthProfile(ArrayList<CoiledCoil> coilList) {
		HashMap<Integer, Integer> resultMap = new HashMap<Integer, Integer>();
		for(CoiledCoil cc: coilList) {
			int count = (resultMap.containsKey(cc.length())) ? resultMap.get(cc.length()) : 0;
			resultMap.put(cc.length(), count + 1);
		}
		return resultMap;
	}


	/**
	 * Takes a list of CoiledCoils
	 * Return a HashMap from a length to a list of CoiledCoils with that length.
	 * @param coilList
	 * @return
	 */
	public static HashMap<Integer, ArrayList<CoiledCoil>> lengthSplitting(ArrayList<CoiledCoil> coilList) {
		HashMap<Integer, ArrayList<CoiledCoil>> resultMap = new HashMap<Integer, ArrayList<CoiledCoil>>();
		for(CoiledCoil cc: coilList) {
			if (resultMap.containsKey(cc.length())) {
				resultMap.get(cc.length()).add(cc);
			} else {
				ArrayList<CoiledCoil> ccList = new ArrayList<CoiledCoil>();
				ccList.add(cc);
				resultMap.put(cc.length(), ccList);
			}
		}
		return resultMap;
	}

	/**
	 * Takes a list of CoiledCoils and returns a map from residue location (1st, 2nd etc.)
	 * to average charge of all residues at the position.
	 * @param coilList
	 * @return
	 */
	public static HashMap<Integer, Double> chargeByPosition(ArrayList<CoiledCoil> coilList) {
		HashMap<Integer, Double> sum = new HashMap<Integer, Double>();
		HashMap<Integer, Double> countMap = new HashMap<Integer, Double>();
		for(CoiledCoil cc: coilList) {
			for(int i=0; i < cc.residueList().size(); i++) {
				double charge = cc.residueList().get(i).getAA().charge();
				if (countMap.containsKey(i)) {
					double currentSum = sum.get(i);
					double currentCount = countMap.get(i);
					sum.put(i, currentSum + charge);
					countMap.put(i,currentCount + 1);	
				} else {
					sum.put(i, (double) charge);
					countMap.put(i, (double) 1);
				}

			}
		}
		HashMap<Integer, Double> average = new HashMap<Integer, Double>();
		for (int i: sum.keySet()) {
			average.put(i, sum.get(i) / countMap.get(i));
		}
		return average;
	}

	public static boolean test(ArrayList<CoiledCoil> coilList, CoiledCoil cc) {
		return false;
	}

	//	public static double test(ArrayList<CoiledCoil> coilList) {
	//		for(CoiledCoil cc: coilList) {
	////			int correct
	//			ArrayList<CoiledCoil> coilListMod = (ArrayList<CoiledCoil>) coilList.clone();
	//			coilListMod.remove(cc);
	//			test(coilListMod, cc);
	//		}
	//	}
}
