package coilPred;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.HashMap;

import sequences.AminoAcid;
import sequences.CoilState;
import sequences.CoilType;
import sequences.Heptad;

public class Viterbi {
	/**
	 * forward_viterbi computes the forward algorithm and the viterbi algorithm
	 * with optimizations for the MulticoilHMM case. 
	 * @param obs
	 * @return HashMap<CoilState,Object[]>.
	 * Object[0] = forward prob of the state
	 * Object[1] = viterbi path through the state
	 * Object[2] = probability of the viterbi path
	 */
	public static HashMap<CoilState,Object[]> forward_viterbi(CoilHMM hmm, ArrayList<AminoAcid> obs) {
		ArrayList<CoilState> HMM_states = hmm.HMM_states;
		ArrayList<CoilState> start_states = hmm.start_states;
		HashMap<CoilState,ArrayList<CoilState>> trans = hmm.trans;
		HashMap<CoilState,ArrayList<CoilState>> reverse_trans = hmm.reverse_trans;
		//T is the old viterbi information:
		HashMap<CoilState,Object[]> T = new HashMap<CoilState, Object[]>();
		//Initialize T for each of the states:
		for (CoilState state: HMM_states) {
			T.put(state, new Object[] {0.0, new ArrayList<CoilState>(),0.0});
		}
		for (CoilState state: start_states) {
			T.put(state, new Object[] {1.0,new ArrayList<CoilState>(),1.0});
		}
		//Do the first step separately:
		for (CoilState state: HMM_states) {
			if (start_states.contains(state)) {
				double val_max = ((Double) T.get(state)[0]);
				double p = ((Double) T.get(state)[2]);
				p *= get_emit_p(hmm,state,0,obs);
				val_max *= get_emit_p(hmm,state,0,obs);
				T.put(state, new Object[] {val_max, new ArrayList<CoilState>(), p});
			}
		}

		for (int i=1; i<obs.size();i++) {
			AminoAcid output = obs.get(i);
			HashMap<CoilState, Object[]> U = new HashMap<CoilState, Object[]>();
			for (CoilState new_state: HMM_states) {
				//The new state will have
				double prob = 0;
				CoilState arg_max = null;
				ArrayList<CoilState> v_path = null;
				double val_max = 0;
				//Because all transition probabilities are 1, we just need to 
				//find max(prob of old_states) or sum(prob of old_states)
				//and then only multiply by our emission probability at the end:
				for (CoilState old_state: reverse_trans.get(new_state)) {
					//Only compute things if the old_state had positive v_prob:
					double old_prob = ((Double) T.get(old_state)[2]);
					//TODO: check to make sure below if statement doesn't fire by accident:
					if (old_prob > 0) {
						prob += old_prob;
						if (old_prob > val_max) {
							val_max = old_prob;
							arg_max = old_state;
						}
					}
				}
				//Now multiply prob, val_max by our emission probability, but
				//only if the state has positive probability:
				if (val_max > 0) {
					double p = get_emit_p(hmm,new_state,i,obs);
					prob *= p;
					val_max *= p;
					//Not sure why we have to cast twice here... it's a little weird
					v_path = (ArrayList<CoilState>) ((ArrayList<CoilState>) T.get(arg_max)[1]).clone();
					v_path.add(arg_max);
					U.put(new_state, new Object[] {prob,v_path,val_max});
				} else {
					U.put(new_state, new Object[] {0.0,new ArrayList<CoilState>(),0.0});
				}
			}
			T = U;
		}
		return T;
	}

	public static HashMap<CoilState,double[]> forward_backward(CoilHMM hmm, ArrayList<AminoAcid> obs) {
		HashMap<CoilState,double[]> forward = forward(hmm,obs);
		HashMap<CoilState,double[]> backward = backward(hmm,obs);
		HashMap<CoilState,double[]> f_b = new HashMap<CoilState, double[]>();
		for(CoilState cs: hmm.HMM_states) {
			f_b.put(cs, new double[obs.size()]);
		}
		for(int i=0;i<obs.size();i++) {
			double sum = 0.0;
			for(CoilState cs: hmm.HMM_states) {
				//If statement here so we don't ask for invalid emit
				//statements
				if(forward.get(cs)[i]*backward.get(cs)[i]>0) {
					f_b.get(cs)[i] = forward.get(cs)[i]*backward.get(cs)[i]/ get_emit_p(hmm,cs,i,obs);
					sum += f_b.get(cs)[i];
				}
			}
			//Normalize to 1
			for(CoilState cs: hmm.HMM_states) {
				f_b.get(cs)[i] /= sum;
			}
		}
		return f_b;
	}

	public static HashMap<CoilState,double[]> forward(CoilHMM hmm, ArrayList<AminoAcid> obs) {
		//Note forward[i] includes the emission probability at location i.
		//backward[i] does not include the emission probability at location i,
		//so when you multiply them you will get the total probability at location i
		ArrayList<CoilState> HMM_states = hmm.HMM_states;
		ArrayList<CoilState> start_states = hmm.start_states;
		HashMap<CoilState,ArrayList<CoilState>> reverse_trans = hmm.reverse_trans;
		//T is the old forward information:
		HashMap<CoilState,double[]> T = new HashMap<CoilState, double[]>();
		//Initialize T for each of the states:
		for (CoilState state: HMM_states) {
			T.put(state, new double[obs.size()]);
		}
		
		for (CoilState state: start_states) {
			T.get(state)[0] = 1.0;
		}
		//Do the first step separately:
		for (CoilState state: HMM_states) {
			if (start_states.contains(state)) {
				double p = T.get(state)[0];
				p *= get_emit_p(hmm,state,0,obs);
				T.get(state)[0] = p;
			}
		}

		for (int i=1; i<obs.size();i++) {
			for (CoilState new_state: HMM_states) {
				//The new state will have
				double prob = 0;
				//Because all transition probabilities are 1, we just need to 
				//find sum(prob of old_states)
				//and then only multiply by our emission probability at the end:
				for (CoilState old_state: reverse_trans.get(new_state)) {
					//Only compute things if the old_state had positive v_prob:
					double[] old_state_probs = T.get(old_state);
					double old_prob = old_state_probs[i-1];
					//TODO: check to make sure below if statement doesn't fire by accident:
					if (old_prob > 0) {
						prob += old_prob;
					}
				}
				//Now multiply prob, val_max by our emission probability, but
				//only if the state has positive probability:
				if (prob > 0) {
					double p = get_emit_p(hmm,new_state,i,obs);
					prob *= p;
					T.get(new_state)[i] = prob;
				} else {
					T.get(new_state)[i] = 0.0;
				}
			}
			
			//Normalize:
			double sum = 0.0;
			for(CoilState cs: hmm.HMM_states) {
				//If statement here so we don't ask for invalid emit
				//statements
				if(T.get(cs)[i]>0){
					sum += T.get(cs)[i];
				}
			}
			for(CoilState cs: hmm.HMM_states) {
				T.get(cs)[i] /= sum;
			}
		}
		return T;		
	}

	public static HashMap<CoilState,double[]> backward(CoilHMM hmm, ArrayList<AminoAcid> obs) {
		//Note forward[i] and backward[i] both include the emission probability at location i.
		//so forward[i]*backward[i] / emission_probability(i) will give the total probability
		//at i.
		ArrayList<CoilState> HMM_states = hmm.HMM_states;
		//Right now all states are set to be end_states, but we can replace this with
		//some end_states subset if we want to.
		ArrayList<CoilState> end_states = hmm.end_states;
		HashMap<CoilState,ArrayList<CoilState>> trans = hmm.trans;
		HashMap<CoilState,double[]> T = new HashMap<CoilState, double[]>();
		//Initialize T for each of the states:
		for (CoilState state: HMM_states) {
			T.put(state, new double[obs.size()]);
		}
		for (CoilState state: end_states) {
			T.get(state)[obs.size()-1] = 1.0;
		}
		//Do the first step (last position) separately:
		for (CoilState state: HMM_states) {
			if (end_states.contains(state)) {
				//Get the last probability
				double p = T.get(state)[obs.size()-1];
				int last_index = obs.size()-1;
				p *= get_emit_p(hmm,state,last_index,obs);
				T.get(state)[obs.size()-1] = p;
			}
		}

		for (int i=obs.size()-2; i>=0; i--) {
			for (CoilState new_state: HMM_states) {
				//The new state will have
				double prob = 0;
				//Because all transition probabilities are 1, we just need to 
				//find max(prob of old_states) or sum(prob of old_states)
				//and then only multiply by our emission probability at the end:
				for (CoilState old_state: trans.get(new_state)) {
					//Only compute things if the old_state had positive v_prob:
					double[] state_probs = T.get(old_state);
					double old_prob = state_probs[i+1];
					//TODO: check to make sure below if statement doesn't fire by accident:
					if (old_prob > 0) {
						prob += old_prob;
					}
				}
				//Now multiply prob, val_max by our emission probability, but
				//only if the state has positive probability:
				//Second part of the conditional so we don't query about
				//a state that can't be in position i
				if (prob > 0 && i >= new_state.position().toInt()-1) {
					double p = get_emit_p(hmm,new_state,i,obs);
					prob *= p;
					T.get(new_state)[i] = prob;
				} else {
					T.get(new_state)[i] = 0.0;
				}
			}
			//Normalize:
			double sum = 0.0;
			for(CoilState cs: hmm.HMM_states) {
				//If statement here so we don't ask for invalid emit
				//statements
				if(T.get(cs)[i]>0){
					sum += T.get(cs)[i];
				}
			}
			for(CoilState cs: hmm.HMM_states) {
				T.get(cs)[i] /= sum;
			}
			
		}
		return T;
	}


	public static double get_emit_p(CoilHMM hmm, CoilState state, int i, ArrayList<AminoAcid> obs) {
		HashMap<Integer, EnumMap<Heptad,EnumMap<AminoAcid,EnumMap<AminoAcid,Double>>>> dimer_corr = hmm.dimer_corr;
		HashMap<Integer, EnumMap<Heptad,EnumMap<AminoAcid,EnumMap<AminoAcid,Double>>>> trimer_corr = hmm.trimer_corr;

		AminoAcid aa = obs.get(i);
		int pos = state.pos.toInt();
		Heptad h = state.hept;
		
		//assert i - pos + 1 >= 0;
		if (pos > i+1) {
			return 0;
		}
		double p = hmm.emit_p.get(state).get(aa);
		//The non coil correlations are in here - every position needs a multiplication!!!
		if (state.type == CoilType.DIMER) {
			switch (pos) {
			case 9:
			case 8:
				p *= dimer_corr.get(7).get(h).get(obs.get(i-7)).get(aa);
			case 7:
				p *= dimer_corr.get(6).get(h).get(obs.get(i-6)).get(aa);
			case 6:
				p *= dimer_corr.get(5).get(h).get(obs.get(i-5)).get(aa);
			case 5:
				p *= dimer_corr.get(4).get(h).get(obs.get(i-4)).get(aa);
			case 4:
				p *= dimer_corr.get(3).get(h).get(obs.get(i-3)).get(aa);
			case 3:
				p *= dimer_corr.get(2).get(h).get(obs.get(i-2)).get(aa);
			case 2:
				p *= dimer_corr.get(1).get(h).get(obs.get(i-1)).get(aa);
			case 1:
			}
		} else if (state.type == CoilType.TRIMER) {
			switch (pos) {
			case 9:
			case 8:
				p *= trimer_corr.get(7).get(h).get(obs.get(i-7)).get(aa);
			case 7:
				p *= trimer_corr.get(6).get(h).get(obs.get(i-6)).get(aa);
			case 6:
				p *= trimer_corr.get(5).get(h).get(obs.get(i-5)).get(aa);
			case 5:
				p *= trimer_corr.get(4).get(h).get(obs.get(i-4)).get(aa);
			case 4:
				p *= trimer_corr.get(3).get(h).get(obs.get(i-3)).get(aa);
			case 3:
				p *= trimer_corr.get(2).get(h).get(obs.get(i-2)).get(aa);
			case 2:
				p *= trimer_corr.get(1).get(h).get(obs.get(i-1)).get(aa);
			case 1:
			}
		} else {
			//state.type == NONE
		}
		return p;
	}
}