package sequences;

import java.util.ArrayList;
import java.util.List;

public class Sequence {
	protected ArrayList<Residue> residueList;
	public String name;
	
	public Sequence(ArrayList<Residue> residueList) {
		this.residueList = residueList;
	}

	public Sequence() {
		residueList = new ArrayList<Residue>();
	}

	public Sequence(CoilType type, String residues, String heptads) {
		residueList = new ArrayList<Residue>();
		for (int i = 0; i<residues.length()-1; i++) {
			AminoAcid aa = AminoAcid.valueOf("" + residues.charAt(i));
			if(heptads.charAt(i) == '.') {
				residueList.add(new NonCoilResidue(aa));
			} else {
				Heptad h = Heptad.valueOf("" + heptads.charAt(i));
				residueList.add(new CoilResidue(type, aa, h));
			}
		}
	}

	public Sequence(String name, String residues) {
		this(residues);
		this.name = name;
	}
	
	public Sequence(String residues) {
		residueList = new ArrayList<Residue>();
		for (int i = 0; i<residues.length(); i++) {
			AminoAcid aa;
			//convert U -> C
			//		  O -> K
			switch (residues.charAt(i)) {
			case 'O':
				aa = AminoAcid.K;
				break;
			case 'U':
				aa = AminoAcid.C;
				break;
			default:
				aa = AminoAcid.valueOf("" + residues.charAt(i));
				break;
			}
			residueList.add(new NonCoilResidue(aa));
		}
	}

	public Sequence add(Residue r) {
		residueList.add(r);
		return this;
	}

	public Sequence add(List<Residue> residueList) {
		this.residueList.addAll(residueList);
		return this;
	}

	public ArrayList<CoiledCoil> toCoils() {
		ArrayList<CoiledCoil> result = new ArrayList<CoiledCoil>();
		CoiledCoil currentCoil = new CoiledCoil();
		for(Residue r: residueList) {
			if(r instanceof CoilResidue) {
				currentCoil.add(((CoilResidue) r).type(), r.getAA(), ((CoilResidue) r).getHept());
			} else if(currentCoil.length() == 0){
			} else {
				result.add(currentCoil);
				currentCoil = new CoiledCoil();
			}
		}
		if(currentCoil.length() != 0) {
			result.add(currentCoil);
		}
		return result;
	}
	
	public CoiledCoil toCoil() {
		ArrayList<CoiledCoil> out = toCoils();
		assert out.size() == 1;
		return toCoils().get(0);
	}

	public ArrayList<CoiledCoil> toCoils(int start, int end) {
		ArrayList<CoiledCoil> result = new ArrayList<CoiledCoil>();
		CoiledCoil currentCoil = new CoiledCoil();
		for(int i=0; i<length(); i++) {
			Residue r = residueList.get(i);
			if(r instanceof CoilResidue) {
				currentCoil.add(((CoilResidue) r).type(), r.getAA(), ((CoilResidue) r).getHept());
			} else if(i>=end) {
				result.add(currentCoil);
				break;
			} else if(i<start) {
				currentCoil = new CoiledCoil();
			} else if(currentCoil.length() == 0){
			} else {
				result.add(currentCoil);
				currentCoil = new CoiledCoil();
			}
		}
		if(currentCoil.length() != 0) {
			result.add(currentCoil);
		}
		return result;
	}


	public int firstCoilRes() {
		for(int i=0;i<residueList.size();i++) {
			if(residueList.get(i) instanceof CoilResidue) {
				return i;
			}
		}
		return residueList.size();
	}

	public int lastCoilRes() {
		for(int i=residueList.size()-1;i>=0;i--) {
			if(residueList.get(i) instanceof CoilResidue) {
				return i+1;
			}
		}
		return 0;
	}

	/**
	 * Cuts the sequence up by coiledcoil (with a gap of 7 residues on
	 * either side of the cc) and then returns a list of the results
	 * @return
	 * Sequence of residues that are not near any CCs
	 */
	public Sequence toNonCoilResidues() {
		Sequence result = new Sequence();
		for(int i=0; i<residueList.size(); i++) {
			if(i<7) {
				if(!(residueList.get(i+7) instanceof CoilResidue)) {
					result.add(residueList.get(i));					
				}
			} else if (i+7>=residueList.size()) {
				if(!(residueList.get(i-7) instanceof CoilResidue)) {
					result.add(residueList.get(i));
				}
			} else if(!(residueList.get(i-7) instanceof CoilResidue) && !(residueList.get(i+7) instanceof CoilResidue)) {
				result.add(residueList.get(i));
			}
		}
		return result;
	}

	public Sequence toNearNonCoilResidues() {
		Sequence result = new Sequence();
		//One run through the residues forward, and then one run through backward
		for(int i=0; i<residueList.size()-1; i++) {
			if((i==0 || !(residueList.get(i-1) instanceof CoilResidue)) && (residueList.get(i) instanceof CoilResidue)) {
				for(int j=Math.max(i-100, 0);j<i;j++) {
					result.add(residueList.get(j));
				}
				break;
			}
		}
		for(int i=residueList.size()-1;i>0;i--) {
			if (((i==residueList.size()-1) || !(residueList.get(i+1) instanceof CoilResidue)) && (residueList.get(i) instanceof CoilResidue)) {
				for(int j=Math.min(i+100, residueList.size()-1);j>i;j--) {
					result.add(residueList.get(j));
				}
				break;
			}
		}
		return result;
	}

	public Sequence toNearNonCoilResidues(int startingPoint, int length) {
		Sequence result = new Sequence();
		result.add(residueList.subList(Math.max(startingPoint-100, 0), startingPoint));
		result.add(residueList.subList(startingPoint+length, Math.min(startingPoint+length+100, residueList.size())));
		return result;
	}

	public Sequence toNonCoilResidues(int startingPoint, int length) {
		Sequence result = new Sequence();
		if(startingPoint>7) {
			result.add(residueList.subList(0, startingPoint));
		}
		if(startingPoint + length + 7 < residueList.size()) {
			result.add(residueList.subList(startingPoint+length+7, residueList.size()));
		}
		return result;
	}

	public CoiledCoil toCoiledCoil(int startingPoint, Heptad hept, int length, CoilType type) {
		CoiledCoil result = new CoiledCoil();
		Heptad currentHept = hept;
		for(int i=0; i<length; i++) {
			result.add(type, residueList.get(startingPoint+i).getAA(), currentHept);
			currentHept = currentHept.prev(6);
		}
		return result;
	}

	public Residue getResidue(int i) {
		return residueList.get(i);
	}

	public AminoAcid getAA(int i) {
		return residueList.get(i).getAA();
	}

	public int length() {
		return residueList.size();
	}

	public ArrayList<Residue> getResidueList() {
		return this.residueList;
	}

	@Override
	public String toString() {
		String residues = "";
		String coils = "";
		String type = "";
		for(Residue r: residueList) {
			residues += r.getAA().toString();
			if(r instanceof CoilResidue) {
				coils += ((CoilResidue) r).getHept();
				type += ((CoilResidue) r).type();
			} else {
				coils += "-";
				type += ".";
			}
		}
		return residues + "\n" + coils + "\n" + type;
	}

	public ArrayList<AminoAcid> toObservedArray() {
		ArrayList<AminoAcid> result = new ArrayList<AminoAcid>();
		for(int i=0; i<length(); i++) {
			result.add(residueList.get(i).getAA());
		}
		return result;
	}
}
