package tables;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.util.HashMap;

import sequences.AminoAcid;
import sequences.Heptad;
/**
 * For each triple of first heptad, a second heptad and distance between them,
 * contains a table of probabilities of each pair of AminoAcids occurring at 
 * those two locations.
 * For example: Getting heptad1 = 'a', heptad2 = 'c', i = 2, AminoAcid1, AminoAcid2
 * will return the probability that, for any two residues located two apart which
 * are labeled 'a' and 'c' respectively (that is, there are no gaps between them)
 * AminoAcid1 occurs at heptad1 and AminoAcid2 occurs at heptad2.
 * @author jtrigg
 *
 */
public class HeptadPairwiseProbTable {
	HashMap<Heptad, HashMap<Heptad, PairwiseProbTable>> table = 
		new HashMap<Heptad, HashMap<Heptad,PairwiseProbTable>>();
	ProbMap map;
	boolean isCustom = false;
	
	public HeptadPairwiseProbTable(int[] diffVals) {
		for(Heptad h1: Heptad.values()) {
			HashMap<Heptad, PairwiseProbTable> subTable = new HashMap<Heptad, PairwiseProbTable>();
			for(Heptad h2: Heptad.values()) {
				subTable.put(h2, new PairwiseProbTable(diffVals));
			}
			table.put(h1,subTable);	
		}
	}
	
	public HeptadPairwiseProbTable(BufferedReader b) throws IOException {
		String s;
		s = b.readLine();
		assert s.equals("HeptadPairwiseProbTable");
		for(Heptad h1: Heptad.values()) {
			s = b.readLine();
			assert s.equals(h1.toString());
			HashMap<Heptad, PairwiseProbTable> subTable = new HashMap<Heptad, PairwiseProbTable>();
			for(Heptad h2: Heptad.values()) {
				s = b.readLine();
				assert s.equals(h2.toString());
				subTable.put(h2, new PairwiseProbTable(b));
			}
			table.put(h1, subTable);
		}
	}
	
	public void writeObject(BufferedWriter b) throws IOException {
	b.write("HeptadPairwiseProbTable\n");
		for(Heptad h1: Heptad.values()) {
			b.write(h1.toString()+"\n");
			for(Heptad h2: Heptad.values()) {
				b.write(h2.toString()+"\n");
				table.get(h1).get(h2).writeObject(b);
			}
		}
	}
	
	public void setProbMap(ProbMap pm) {
		map = pm;
		isCustom = true;
	}
	
	public void add(Heptad h1, Heptad h2, int diff, AminoAcid aa1,
			AminoAcid aa2, double num) {
		table.get(h1).get(h2).add(diff, aa1, aa2, num);
	}
	
	public void add(Heptad h1, Heptad h2, PairwiseProbTable newTable) {
		this.table.get(h1).get(h2).put(newTable);
	}
	
	public double get(Heptad h1, Heptad h2, int diff, AminoAcid aa1,
			AminoAcid aa2) {
		double prob = table.get(h1).get(h2).get(diff, aa1, aa2);
		if(isCustom) {
			//TODO clean up this var
			double tempProb = map.map(prob);
			return map.map(prob);
		} else {
			return prob;
		}
	}
	
	@Override
	public String toString() {
		String result = "";
		for(Heptad h: table.keySet()) {
			for(Heptad h2: table.get(h).keySet()) {
				result += "\n" + h + " " + h2 + " " + table.get(h).get(h2);
			}
		}
		return result;
	}
}