package tables;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashMap;

import sequences.AminoAcid;
import sequences.Heptad;

/**
 * For every heptad, contains a probability table corresponding to 
 * the probability of each AminoAcid occurring at that heptad.
 * @author jtrigg
 *
 */
public class HeptadResProbTable {
	HashMap<Heptad, ResProbTable> table = new HashMap<Heptad, ResProbTable>();
	ProbMap map;
	boolean isCustom = false;
	
	public HeptadResProbTable() {
		for(Heptad h: Heptad.values()) {
			ResProbTable subTable = new ResProbTable();
			table.put(h, subTable);
		}
	}
	
	/**
	 * @throws IOException 
	 */
	public HeptadResProbTable(BufferedReader b) throws IOException{
		String s;
		s = b.readLine();
		assert s.equals("HeptadResProbTable");
		for(Heptad h: Heptad.values()) {
			s = b.readLine();
			assert s.equals(h.toString());
			table.put(h, new ResProbTable(b));
		}
	}

	public void writeObject(BufferedWriter b) throws IOException {
		b.write("HeptadResProbTable\n");
		for(Heptad h: Heptad.values()) {
			b.write(h.toString()+"\n");
			table.get(h).writeObject(b);
		}
	}
	
	/**
	 * Write the HeptadResProbTable to an output stream
	 * @param pm
	 * @throws IOException 
	 */
	public void toFile(FileWriter out) throws IOException {
		table = new HashMap<Heptad,ResProbTable>();
		for(Heptad h: Heptad.values()) {
			for(AminoAcid aa: AminoAcid.values()) {
				out.write(String.valueOf(table.get(h).get(aa)));
			}
		}
	}
	
	public void setProbMap(ProbMap pm) {
		map = pm;
		isCustom = true;
	}
	
	public void put(Heptad h, ResProbTable subTable) {
		this.table.put(h, subTable);
	}

	public void add(Heptad h, AminoAcid aa, double num) {
		table.get(h).add(aa,num);			
	}
	
	public void add(HeptadResProbTable addTable) {
		for(Heptad hept: Heptad.values()) {
			table.get(hept).add(addTable.get(hept));
		}
	}
	
	public static HeptadResProbTable combine(HeptadResProbTable table1, double weight1, HeptadResProbTable table2, double weight2) {
		HeptadResProbTable result = new HeptadResProbTable();
		for(Heptad h: Heptad.values()) {
			result.put(h, ResProbTable.combine(table1.get(h), weight1, table2.get(h), weight2));
		}
		return result;
	}
	
	public ResProbTable get(Heptad h) {
		return table.get(h);
	}
	
	public double get(Heptad h, AminoAcid aa) {
		double prob = table.get(h).get(aa);
		if(isCustom) {
			return map.map(prob);
		} else {
			return prob;
		}
	}
		
	@Override
	public String toString() {
		String result = "";
		for (Heptad h: Heptad.values()) {
			result += h + " " + table.get(h).toString() + "\n";
		}
		return result;
	}
}
