package tables;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;

import sequences.AminoAcid;

/**
 * Contains a table of mapping triples of integer distances, 
 * AminoAcid1 and AminoAcid2 to probabilities. The probability
 * represents the chance that, for any two residues of the given
 * distance, the first one will be AminoAcid1 and the second
 * AminoAcid2
 * 
 * @author jtrigg
 *
 */
public class PairwiseProbTable {
	private HashMap<Integer, HashMap<AminoAcid, HashMap<AminoAcid, Double>>> table = new HashMap<Integer, HashMap<AminoAcid,HashMap<AminoAcid,Double>>>();
	HashMap<Integer, Double> countTable = new HashMap<Integer, Double>();
	public PairwiseProbTable(int i) {
		countTable.put(i,0.0);
		HashMap<AminoAcid, HashMap<AminoAcid, Double>> newSubMap = new HashMap<AminoAcid, HashMap<AminoAcid,Double>>();
		for(AminoAcid aa1: AminoAcid.standardValues()) {
			HashMap<AminoAcid, Double> newSubSubMap = new HashMap<AminoAcid, Double>();
			for(AminoAcid aa2: AminoAcid.standardValues()) {
				newSubSubMap.put(aa2, 0.0);
			}
			newSubMap.put(aa1, newSubSubMap);
		}
		table.put(i, newSubMap);		
	}

	public PairwiseProbTable(int[] intRange) {
		table = new HashMap<Integer, HashMap<AminoAcid, HashMap<AminoAcid, Double>>>();
		for(int i: intRange) {
			HashMap<AminoAcid, HashMap<AminoAcid, Double>> newSubMap = new HashMap<AminoAcid, HashMap<AminoAcid,Double>>();
			countTable.put(i, 0.0);
			for(AminoAcid aa1: AminoAcid.standardValues()) {
				HashMap<AminoAcid, Double> newSubSubMap = new HashMap<AminoAcid, Double>();
				for(AminoAcid aa2: AminoAcid.standardValues()) {
					newSubSubMap.put(aa2, 0.0);
				}
				newSubMap.put(aa1, newSubSubMap);
			}
			table.put(i, newSubMap);
		}
	}

	public PairwiseProbTable(BufferedReader b) throws IOException {
		String s;
		s = b.readLine();
		assert s.equals("PairwiseProbTable");
		s = b.readLine();
		assert s.equals("CountTable");
		s = b.readLine();
		while(!s.equals("Table")) {
			int i = Integer.parseInt(s);
			s = b.readLine();
			double d = Double.parseDouble(s);
			countTable.put(i, d);
			s = b.readLine();
		}
		List<Integer> list = new ArrayList<Integer>(countTable.keySet());
		Collections.sort(list);
		for(int i: list) {
			s = b.readLine();
			assert s.equals(i+"");
			HashMap<AminoAcid, HashMap<AminoAcid, Double>> newSubMap = new HashMap<AminoAcid, HashMap<AminoAcid,Double>>();				
			for(AminoAcid aa1: AminoAcid.standardValues()) {
				HashMap<AminoAcid, Double> newSubSubMap = new HashMap<AminoAcid, Double>();
				s = b.readLine();
				assert aa1.toString().equals(s);
				for(AminoAcid aa2: AminoAcid.standardValues()) {
					s = b.readLine();
					assert aa2.toString().equals(s);
					s = b.readLine();
					double d = Double.parseDouble(s);
					newSubSubMap.put(aa2, d);
				}
				newSubMap.put(aa1, newSubSubMap);
			}
			table.put(i, newSubMap);
		}
	}

	public void writeObject(BufferedWriter b) throws IOException {
		b.write("PairwiseProbTable\n");
		b.write("CountTable\n");
		for(int i: countTable.keySet()){
			b.write(i+"\n");
			b.write(countTable.get(i)+"\n");
		}
		b.write("Table\n");
		for(int i: table.keySet()) {
			b.write(i+"\n");
			for(AminoAcid aa1: AminoAcid.standardValues()) {
				b.write(aa1.toString()+"\n");
				for(AminoAcid aa2: AminoAcid.standardValues()) {
					b.write(aa2.toString()+"\n");
					b.write(table.get(i).get(aa1).get(aa2).toString()+"\n");
				}
			}
		}
		b.flush();
	}

	//public void readObject

	public void add(int i, AminoAcid aa1, AminoAcid aa2, double num) {
		table.get(i).get(aa1).put(aa2, table.get(i).get(aa1).get(aa2) + num);
		countTable.put(i, countTable.get(i) + num);
	}

	public void put(PairwiseProbTable newTable) {
		for(int i:newTable.table.keySet()) {
			table.put(i, newTable.table.get(i));
			countTable.put(i, newTable.countTable.get(i));
		}
	}

	public double get(int i, AminoAcid aa1, AminoAcid aa2) {
		if(aa1 == AminoAcid.B) {
			return .5 * (get(i, AminoAcid.N,aa2) + get(i, AminoAcid.D,aa2));
		} else if (aa1 == AminoAcid.Z) {
			return .5 * (get(i, AminoAcid.E,aa2) + get(i, AminoAcid.Q,aa2));
		} else if (aa1 == AminoAcid.J) {
			return .5 * (get(i, AminoAcid.I,aa2) + get(i, AminoAcid.L,aa2));
		} else if (aa1 == AminoAcid.X) {
			double totalProb = 0.0;
			for(AminoAcid aaPrime: AminoAcid.standardValues()) {
				totalProb += get(i, aaPrime, aa2);
			}
			return .05 * totalProb;
		} else if (aa2 == AminoAcid.B) {
			return .5 * (get(i, aa1, AminoAcid.N) + get(i, aa1, AminoAcid.D));
		} else if (aa2 == AminoAcid.Z) {
			return .5 * (get(i, aa1, AminoAcid.E) + get(i, aa1, AminoAcid.Q));
		} else if (aa2 == AminoAcid.J) {
			return .5 * (get(i, aa1, AminoAcid.I) + get(i, aa1, AminoAcid.L));
		} else if (aa2 == AminoAcid.X) {
			double totalProb = 0.0;
			for(AminoAcid aaPrime: AminoAcid.standardValues()) {
				totalProb += get(i, aa1, aaPrime);
			}
			return .05 * totalProb;
		} else {
			return table.get(i).get(aa1).get(aa2) / countTable.get(i);
		}
	}

	@Override
	public String toString() {
		String result = "";
		for(int i: table.keySet()) {
			result += "\n" + i;
			for(AminoAcid aa1: table.get(i).keySet()) {
				for(AminoAcid aa2: table.get(i).get(aa1).keySet()) {
					result += "\n" + aa1 + " " + aa2 + " " + table.get(i).get(aa1).get(aa2);
				}
			}
		}
		return result;
	}
}
