package tables;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

import sequences.AminoAcid;
/**
 * Maps AminoAcids to probability of occurrance.
 * @author jtrigg
 *
 */
public class ResProbTable {
	HashMap<AminoAcid,Double> table = new HashMap<AminoAcid, Double>();
	double count;
	
	public ResProbTable() {
		for(AminoAcid aa: AminoAcid.standardValues()) {
			table.put(aa,0.0);
		}
		count = 0.0;
	}
	
	public ResProbTable(HashMap<AminoAcid, Double> table, int entries) {
		this.table = table;
		this.count = entries;
	}
	
	public ResProbTable(BufferedReader b) throws IOException {
		String s;
		s = b.readLine();
		assert s.equals("ResProbTable");
		s = b.readLine();
		assert s.equals("Count");
		s = b.readLine();
		count = Double.parseDouble(s);
		for(AminoAcid aa: AminoAcid.standardValues()) {
			s = b.readLine();
			assert s.equals(aa.toString());
			Double d = Double.parseDouble(b.readLine());
			table.put(aa, d);
		}
	}
	
	public void writeObject(BufferedWriter b) throws IOException {
		b.write("ResProbTable\n");
		b.write("Count\n");
		b.write(count+"\n");
		for(AminoAcid aa: AminoAcid.standardValues()) {
			b.write(aa.toString()+"\n");
			b.write(table.get(aa).toString()+"\n");
		}
	}
	
	public void add(AminoAcid aa, double num) {
		count += num;
		table.put(aa, table.get(aa) + num);
	}
	
	public void add(ResProbTable addTable) {
		for(AminoAcid aa: AminoAcid.standardValues()) {
			add(aa, addTable.get(aa));
		}
		count += addTable.count;
	}
	
	public static ResProbTable combine(ResProbTable table1, double weight1, ResProbTable table2, double weight2) {
		ResProbTable result = new ResProbTable();
		for(AminoAcid aa: AminoAcid.standardValues()) {
			result.add(aa, (table1.get(aa) * weight1 / table1.count) + (table2.get(aa) * weight2 / table2.count));
		}
		return result;
	}
	
	public double get(AminoAcid aa) {
		if(aa == AminoAcid.B) {
			return .5 * (get(AminoAcid.N) + get(AminoAcid.D));
		} else if (aa == AminoAcid.Z) {
			return .5 * (get(AminoAcid.E) + get(AminoAcid.Q));
		} else if (aa == AminoAcid.J) {
			return .5 * (get(AminoAcid.I) + get(AminoAcid.L));
		} else if (aa == AminoAcid.X) {
			double totalProb = 0.0;
			for(AminoAcid aaPrime: AminoAcid.standardValues()) {
				totalProb += get(aaPrime);
			}
			//TODO get rid of this slow step:
			assert (totalProb < 1.02 && totalProb > .98);
			return .05 * totalProb;
		} else {
		return table.get(aa) / count;
		}
	}
	
	public boolean contains(AminoAcid aa) {
		return table.containsKey(aa);
	}
	
	@Override
	public String toString() {
		return table.toString();
	}
	
	/**
	 * Estimates the variance of the table
	 * @return
	 */
	public double estVar() {
		double estVar = 0.0;
		for (AminoAcid aa: AminoAcid.standardValues()) {
			estVar += (1 - table.get(aa)) * table.get(aa) * count;
		}
		return estVar / (count * (count - 1.0));
	}
	
	public static double estTableArrayVar(ArrayList<ResProbTable> tables) {
		//First compute the average table: 
		double tableListSize = tables.size();
		ResProbTable averageTable = new ResProbTable();
		for(AminoAcid aa: AminoAcid.standardValues()) {
			for(ResProbTable table: tables) {
				averageTable.add(aa, table.get(aa));
			}
		}
		
		//Next compute the sum of the differences between each table
		//and the average table
		double var = 0.0;
		for(ResProbTable table: tables) {
			for(AminoAcid aa: AminoAcid.standardValues()) {
				var += (table.get(aa) - averageTable.get(aa)) * (table.get(aa) - averageTable.get(aa));
			}
		}
		return var / (tableListSize - 1.0);
	}
}
